import collections
import functools
import re

from ... import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

NO_DEFAULT_VALUE = object()


@functools.cache
def get_width(path, default=NO_DEFAULT_VALUE):
    """
    Return displayed width of video file `path`

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if width can't be determined
    """
    try:
        width = utils.mediainfo.lookup(path, ('Video', 'DEFAULT', 'Width'), type=int)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        par = utils.mediainfo.lookup(path, ('Video', 'DEFAULT', 'PixelAspectRatio'), type=float, default=1.0)
        if par > 1.0:
            _log.debug('Display width: %r * %r = %r', width, par, width * par)
            width = int(width * par)
        return width


@functools.cache
def get_height(path, default=NO_DEFAULT_VALUE):
    """
    Return displayed height of video file `path`

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if height can't be determined
    """
    try:
        height = utils.mediainfo.lookup(path, ('Video', 'DEFAULT', 'Height'), type=int)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        par = utils.mediainfo.lookup(path, ('Video', 'DEFAULT', 'PixelAspectRatio'), type=float, default=1.0)
        if par < 1.0:
            _log.debug('Display height: (1 / %r) * %r = %r', par, height, (1 / par) * height)
            height = int((1 / par) * height)
        return height


def get_resolution(path, default=NO_DEFAULT_VALUE):
    """
    Return resolution and scan type of video file `path` as :class:`str` (e.g. "1080p")

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if resolution can't be determined
    """
    try:
        resolution = get_resolution_int(path)
        scan_type = get_scan_type(path)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        return f'{resolution}{scan_type}'


_standard_resolutions = (
    # (width, height, standard resolution)
    # 4:3
    (640, 480, 480),
    (720, 540, 540),
    (768, 576, 576),
    (960, 720, 720),
    (1440, 1080, 1080),
    (2880, 2160, 2160),
    (5760, 4320, 4320),

    # 16:9
    (853, 480, 480),
    (960, 540, 540),
    (1024, 576, 576),
    (1280, 720, 720),
    (1920, 1080, 1080),
    (3840, 2160, 2160),
    (7680, 4320, 4320),

    # 21:9
    (853, 365, 480),
    (960, 411, 540),
    (1024, 438, 576),
    (1280, 548, 720),
    (1920, 822, 1080),
    (3840, 1645, 2160),
    (7680, 3291, 4320),
)

@functools.cache
def get_resolution_int(path, default=NO_DEFAULT_VALUE):
    """
    Return resolution of video file `path` as :class:`int` (e.g. ``1080``)

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if resolution can't be determined
    """
    try:
        actual_width = get_width(path)
        actual_height = get_height(path)
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        # Find distances from actual display width/height to each standard width/height.
        # Categorize them by standard ratio.
        distances = collections.defaultdict(dict)
        for std_width, std_height, std_resolution in _standard_resolutions:
            std_aspect_ratio = round(std_width / std_height, 1)
            w_dist = abs(std_width - actual_width)
            h_dist = abs(std_height - actual_height)
            resolution = (std_width, std_height, std_resolution)
            distances[std_aspect_ratio][w_dist] = distances[std_aspect_ratio][h_dist] = resolution

        # Find the standard aspect ratio that is closest to the given aspect ratio.
        actual_aspect_ratio = round(actual_width / actual_height, 1)
        std_aspect_ratios = tuple(distances)
        closest_std_aspect_ratio = utils.closest_number(actual_aspect_ratio, std_aspect_ratios)

        # Pick the standard resolution with the lowest distance to the given resolution.
        dists = distances[closest_std_aspect_ratio]
        std_width, std_height, std_resolution = sorted(dists.items())[0][1]

        _log.debug(
            'Closest standard resolution: %r x %r [%.1f] -> %r x %r [%.1f] -> %r',
            actual_width, actual_height, actual_aspect_ratio,
            std_width, std_height, closest_std_aspect_ratio,
            std_resolution,
        )
        return std_resolution


def get_scan_type(path):
    """
    Return scan type of video file `path` ("i" for interlaced, "p" for progressive)

    This always defaults to "p" if it cannot be determined.

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :raise ContentError: if scan type can't be determined
    """
    scan_type = utils.mediainfo.lookup(path, ('Video', 'DEFAULT', 'ScanType'), default='p').lower()
    if scan_type in ('interlaced', 'mbaff', 'paff'):
        return 'i'
    else:
        return 'p'


def get_frame_rate(path, default=NO_DEFAULT_VALUE):
    """
    Return frames per second of default video track as :class:`float`

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    return utils.mediainfo.lookup(
        path=path,
        keys=('Video', 'DEFAULT', 'FrameRate'),
        default=default,
        type=float,
    )


def get_bit_depth(path, default=NO_DEFAULT_VALUE):
    """
    Return bit depth of default video track (e.g. ``8`` or ``10``)

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    return utils.mediainfo.lookup(
        path=path,
        keys=('Video', 'DEFAULT', 'BitDepth'),
        default=default,
        type=int,
    )


known_hdr_formats = {
    'DV',
    'HDR10+',
    'HDR10',
    'HDR',
}
"""Set of valid HDR format names"""

def get_hdr_formats(path, default=NO_DEFAULT_VALUE):
    """
    Return sequence of HDR formats e.g. ``("HDR10",)``, ``("DV", "HDR10")``

    The sequence may be empty.

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    def is_dv(video_track):
        return bool(
            # Dolby Vision[ / <more information>]
            re.search(r'^Dolby Vision', video_track.get('HDR_Format', ''))
        )

    def is_hdr10p(video_track):
        return bool(
            # "HDR10+ Profile A" or "HDR10+ Profile B"
            re.search(r'HDR10\+', video_track.get('HDR_Format_Compatibility', ''))
        )

    def is_hdr10(video_track):
        return bool(
            re.search(r'HDR10(?!\+)', video_track.get('HDR_Format_Compatibility', ''))
            or
            re.search(r'BT\.2020', video_track.get('colour_primaries', ''))
        )

    def is_hdr(video_track):
        return bool(
            re.search(r'HDR(?!10)', video_track.get('HDR_Format_Compatibility', ''))
            or
            re.search(r'HDR(?!10)', video_track.get('HDR_Format', ''))
        )

    try:
        video_track = utils.mediainfo.lookup(path, ('Video', 'DEFAULT'))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        hdr_formats = []

        # NOTE: DV and HDR(10)(+) can co-exist.
        if is_dv(video_track):
            hdr_formats.append('DV')

        if is_hdr10p(video_track):
            hdr_formats.append('HDR10+')
        elif is_hdr10(video_track):
            hdr_formats.append('HDR10')
        elif is_hdr(video_track):
            hdr_formats.append('HDR')

        return tuple(hdr_formats)


def is_bt601(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.601 (~SD) video

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist, raise :exc:`~.ContentError` if not
        provided

    :raise ContentError: if anything goes wrong
    """
    try:
        if (
                _is_color_matrix(path, 'BT.601')
                or _is_color_matrix(path, 'BT.470 System B/G')
        ):
            return True
        else:
            # Assume BT.601 if default video is SD.
            # https://rendezvois.github.io/video/screenshots/programs-choices/#color-matrix
            resolution = get_resolution_int(path)
            return resolution < 720
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default

def is_bt709(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.709 (~UHD) video

    See :func:`is_bt601`.
    """
    return _is_color_matrix(path, 'BT.709', default=default)

def is_bt2020(path, default=NO_DEFAULT_VALUE):
    """
    Whether `path` is BT.2020 (~UHD) video

    See :func:`is_bt601`.
    """
    return _is_color_matrix(path, 'BT.2020', default=default)

def _is_color_matrix(path, matrix, default=NO_DEFAULT_VALUE):
    def normalize_matrix(matrix):
        # Remove whitespace and convert to lower case.
        return ''.join(matrix.casefold().split())

    matrix = normalize_matrix(matrix)

    try:
        video_tracks = utils.mediainfo.lookup(path, ('Video',))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        # https://rendezvois.github.io/video/screenshots/programs-choices/#color-matrix
        for track in video_tracks:
            if normalize_matrix(track.get('matrix_coefficients', '')).startswith(matrix):
                return True
        return False


_video_translations = (
    ('x264', {'Encoded_Library_Name': re.compile(r'^x264$')}),
    ('x265', {'Encoded_Library_Name': re.compile(r'^x265$')}),
    ('XviD', {'Encoded_Library_Name': re.compile(r'^XviD$')}),
    ('H.264', {'Format': re.compile(r'^AVC$')}),
    ('H.265', {'Format': re.compile(r'^HEVC$')}),
    ('VP9', {'Format': re.compile(r'^VP9$')}),
    ('VC-1', {'Format': re.compile(r'^VC-1$')}),
    ('MPEG-2', {'Format': re.compile(r'^MPEG Video$')}),
)

@functools.cache
def get_video_format(path, default=NO_DEFAULT_VALUE):
    """
    Return video format of default video track

    Return x264, x265 or XviD if either one is detected.

    :param str path: Path to video file

        For directories, the return value of :func:`find_main_video` is used.

    :param default: Return value if `path` doesn't exist or video format cannot be determined, raise
        :exc:`~.ContentError` if not provided

    :raise ContentError: if anything goes wrong
    """
    try:
        video_track = utils.mediainfo.lookup(path, ('Video', 'DEFAULT'))
    except errors.ContentError as e:
        if default is NO_DEFAULT_VALUE:
            raise e
        else:
            return default
    else:
        for vfmt, regexs in _video_translations:
            for key, regex in regexs.items():
                value = video_track.get(key)
                if value and regex.search(value):
                    _log.debug('Detected video format: %s', vfmt)
                    return vfmt

        if default is NO_DEFAULT_VALUE:
            raise errors.ContentError('Unable to detect video format')
        else:
            _log.debug('Failed to detect video format, falling back to default: %s', default)
            return default



"""
File system helpers
"""

import fnmatch
import functools
import os
import pathlib
import re
import time

import torf

from .. import __project_name__, constants, errors
from . import LazyModule, os_family, types

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = LazyModule(module='natsort', namespace=globals())


OS_SEP_REGEX = re.escape(os.sep)
r"""
:attr:`os.sep` escaped with :func:`re.escape`

On Windows, this should be ``"\\"`` while it should be identical to
:attr:`os.sep` on Unix-likes.
"""


def assert_file_readable(filepath):
    """Raise :exc:`~.ContentError` if `open(filepath)` fails"""
    try:
        open(filepath).close()
    except OSError as e:
        raise errors.ContentError(f'{filepath}: {e.strerror or e}') from e


def assert_dir_usable(path, *, check_writable=True):
    """
    Raise :exc:`~.ContentError` if `dirpath` can not be used as a directory

    :param bool check_writable: Whether to ignore if `path` is not writable
    """
    if not os.path.isdir(path):
        raise errors.ContentError(f'{path}: Not a directory')
    elif not os.access(path, os.R_OK):
        raise errors.ContentError(f'{path}: Not readable')
    elif check_writable and not os.access(path, os.W_OK):
        raise errors.ContentError(f'{path}: Not writable')
    elif not os.access(path, os.X_OK):
        raise errors.ContentError(f'{path}: Not executable')


@functools.cache
def projectdir(content_path, base=None):
    """
    Return path to existing directory in which jobs put their files and cache

    :param str content_path: Path to torrent content
    :param str base: Location of the project directory; defaults to
        :attr:`~.constants.DEFAULT_CACHE_DIRECTORY`

    :raise ContentError: if `content_path` exists and is not a directory or has
        insufficient permissions
    """
    if content_path:
        if not base:
            base = constants.DEFAULT_CACHE_DIRECTORY
        path = os.path.join(base, basename(content_path) + f'.{__project_name__}')
        mkdir(path)
    else:
        path = '.'
    _log.debug('Using project directory: %r', path)
    return path


def ensure_path_in_cache(path, cache_directory):
    """
    Return `path` beneath `cache_directory`

    If `path` points to anywhere inside `cache_directory`, return it
    unmodified. Otherwise, return the :func:`basename` of `path` appended to
    `cache_directory`.
    """
    path = os.path.abspath(path)
    cache_directory = os.path.abspath(cache_directory)
    if os.path.commonpath((path, cache_directory)).startswith(cache_directory):
        # `path` is somewhere beneath `cache_directory`.
        return path
    else:
        # `path` is outside of `cache_directory`.
        return os.path.join(cache_directory, basename(path))


def limit_directory_size(path, max_total_size, min_age=None, max_age=None):
    """
    Delete oldest files (by access time) until maximum size is not exceeded

    Empty files and directories are always deleted.

    :param path: Path to directory
    :param max_total_size: Maximum combined size of all files in `path` and its
        subdirectories
    :param min_age: Preserve files that are younger than this
    :type min_age: Unix timestamp
    :param max_age: Preserve files that are older than this
    :type max_age: Unix timestamp
    """
    @functools.cache
    def cached_file_size(f):
        return file_size(f)

    def combined_size(filepaths):
        return sum(cached_file_size(f) for f in filepaths
                   if os.path.exists(f) and not os.path.islink(f))

    # This should return mtime if file system was mounted with noatime.
    def atime(filepath):
        try:
            return os.stat(filepath, follow_symlinks=False).st_atime
        except OSError:
            return 0

    def get_filepaths(dirpath):
        return file_list(dirpath, min_age=min_age, max_age=max_age, follow_dirlinks=False)

    # How much space do we have to free up?
    filepaths = get_filepaths(path)
    total_size = combined_size(filepaths)
    size_diff = total_size - max_total_size
    if size_diff > 0:
        # Collect old files until they free up enough space when removed
        oldest_files = sorted(filepaths, key=atime)
        files_to_remove = [oldest_files.pop(0)]
        while combined_size(files_to_remove) < size_diff:
            files_to_remove.append(oldest_files.pop(0))

        # Actually remove the files
        for file_to_remove in files_to_remove:
            # It's possible another process already unlinked this file while we
            # were still collecting old files.
            try:
                os.unlink(file_to_remove)
            except FileNotFoundError:
                pass


def prune_empty(path, *, files=False, directories=True):
    """
    Remove empty subdirectories recursively

    :param path: Path to directory
    :param bool files: Whether to prune empty files
    :param bool directories: Whether to prune empty directories

    Dead symbolic links are removed after pruning files and directories.

    If `path` is not a directory, do nothing.
    """
    if not os.path.isdir(path):
        return

    # If multiple instances are running at the same time, one might be pruning
    # empty directories just after another instance has created their (initially
    # empty) project directory. To mitigate this, we only prune old directories
    # (e.g. > 60s).
    def is_older_than(min_age, path, now=time.time()):
        try:
            statinfo = os.stat(path, follow_symlinks=False)
        except OSError:
            return False
        else:
            age = round(now - statinfo.st_mtime)
            return age > min_age

    def raise_error(e, path):
        raise RuntimeError(f'{path}: Failed to prune: {e.strerror or e}')

    # Prune empty files
    if files:
        for dirpath, _dirnames, filenames in os.walk(path, topdown=False, followlinks=False):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                try:
                    if os.path.exists(filepath) and file_size(filepath) <= 0:
                        os.unlink(filepath)
                except OSError as e:
                    raise_error(e, filepath)

    # Prune empty directories
    if directories:
        for dirpath, dirnames, _filenames in os.walk(path, topdown=False, followlinks=False):
            for dirname in dirnames:
                subdirpath = os.path.join(dirpath, dirname)
                try:
                    if (
                            not os.path.islink(subdirpath)
                            and not bool(os.listdir(subdirpath))
                            and is_older_than(60, subdirpath)
                    ):
                        os.rmdir(subdirpath)
                except OSError as e:
                    raise_error(e, subdirpath)

    # Prune dead links
    for dirpath, _dirnames, filenames in os.walk(path, topdown=False, followlinks=False):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            try:
                if os.path.islink(filepath) and not os.path.exists(filepath):
                    os.unlink(filepath)
            except OSError as e:
                raise_error(e, filepath)

    try:
        if directories and not os.listdir(path):
            os.rmdir(path)
    except OSError as e:
        raise_error(e, path)


def mkdir(path):
    """
    Create directory and its parents

    Existing directories are ignored.

    If `path` is empty, no directory is created.

    :raise ContentError: if directory creation fails
    """
    if path:
        try:
            os.makedirs(path, exist_ok=True)
        except OSError as e:
            raise errors.ContentError(f'{path}: {e.strerror or e}') from e
        else:
            assert_dir_usable(path)


def basename(path, parents=0):
    """
    Return last segment in `path`

    Unlike :func:`os.path.basename`, this removes any trailing directory
    separators first.

    >>> os.path.basename('a/b/c/')
    ''
    >>> upsies.utils.basename('a/b/c/')
    'c'

    :param parents: How many levels of parent directories to include in the
        returned path
    """
    basename = os.path.basename(str(path).rstrip(os.sep))
    parent_parts = list(pathlib.Path(str(path)).parent.parts)
    parents_to_include = []
    for _ in range(parents):
        if parent_parts:
            parents_to_include.insert(0, parent_parts.pop(-1))
    return str(pathlib.Path(*parents_to_include).joinpath(basename))


def dirname(path):
    """
    Remove last segment in `path`

    Unlike :func:`os.path.dirname`, this removes any trailing directory
    separators first.

    >>> os.path.dirname('a/b/c/')
    'a/b/c'
    >>> upsies.utils.dirname('a/b/c/')
    'a/b'
    """
    return os.path.dirname(str(path).rstrip(os.sep))


def file_and_parent(path):
    """Return `path`'s filename and parent directoy name if there is one"""
    if os.sep in path:
        return (
            basename(path),
            basename(dirname(path)),
        )
    else:
        return (path,)


def sanitize_filename(filename):
    """
    Replace illegal characters in `filename` with "_"

    Illegal characters include :attr:`os.sep`.
    """
    if os_family() == 'windows':
        illegal_chars = ('<', '>', ':', '"', '/', '\\', '|', '?', '*')
    else:
        illegal_chars = ('/',)
    for char in illegal_chars:
        filename = filename.replace(char, '_')
    return filename


def sanitize_path(path):
    """
    Replace illegal characters in each `path` segment with "_"

    `path` is split at :attr:`os.sep` and the resulting items are passed to
    :func:`sanitize_filename` and joined.

    On Windows, the drive (e.g. "C:") is not sanitized to keep the ":".
    """
    # Get drive from Windows paths (e.g. "C:") because we don't want to sanitize
    # the ":", which is illegal otherwise. If there are no drives (e.g. on
    # Linux), `drive` is empty, so we don't have to handle that case specially.
    drive, path = os.path.splitdrive(str(path))
    segments = str(path).split(os.sep)
    sanitized_segments = [sanitize_filename(segment) for segment in segments]
    return drive + os.sep.join(sanitized_segments)


def tildify_path(path):
    """Return `path` with $HOME replaced by ``~``"""
    home = os.path.expanduser('~')
    return re.sub(rf'^{re.escape(home)}', '~', path)


def file_extension(path, minlen=1, maxlen=None, only=()):
    """
    Return file extension

    Unlike :func:`os.path.splitext`, this function expects the extension to
    consist only of alphanumeric ASCII characters.

    :param str path: Path to file or directory
    :param str minlen: Minimum number of characters in file extension or `None`
        for no limit
    :param str maxlen: Maximum number of characters in file extension or `None`
        for no limit
    :param only: Only return extension if it exists in this sequence
    :type only: sequence of :class:`str`

    :return: file extension
    :rtype: str
    """
    filename = os.path.basename(path)
    match = re.search(rf'\.([a-zA-Z0-9]{_make_minmax(minlen, maxlen)})$', filename)
    if match:
        ext = match.group(1).lower()
        if not only or ext in (str(e).lower() for e in only):
            return ext
    return ''

def strip_extension(path, minlen=1, maxlen=None, only=()):
    """
    Return `path` without file extension

    If `path` doesn't have a file extension, return it as is.

    A file extension consists only of alphanumeric ASCII characters.

    :param str path: Path to file or directory
    :param str minlen: Minimum number of characters in file extension or `None`
        for no limit
    :param str maxlen: Maximum number of characters in file extension or `None`
        for no limit
    :param only: Only strip extension if it exists in this sequence
    :type only: sequence of :class:`str`

    :return: file name without extension
    :rtype: str
    """
    path = str(path)
    match = re.search(rf'^(.*)\.([a-zA-Z0-9]{_make_minmax(minlen, maxlen)})$', path)
    if match:
        name = match.group(1)
        ext = match.group(2).lower()
        if not only or ext in (str(e).lower() for e in only):
            return name
    return path

def _make_minmax(minlen, maxlen):
    if minlen is not None and maxlen is not None:
        return '{' + f'{minlen},{maxlen}' + '}'
    elif minlen is not None:
        return '{' + f'{minlen},' + '}'
    elif maxlen is not None:
        return '{' + f',{maxlen}' + '}'
    else:
        return '*'


def file_size(path):
    """Return file size in bytes or `None`"""
    if not os.path.isdir(path):
        try:
            return os.path.getsize(path)
        except OSError:
            pass
    return None


def path_size(path):
    """
    Combined size of all files beneath `path`

    Returns the same value as :func:`file_size` if `path` is not a directory.

    Symbolic links are ignored.
    """
    filepaths = file_list(path, follow_dirlinks=False)
    return sum(
        file_size(f) or 0
        for f in filepaths
        if not os.path.islink(f)
    )


def listdir(path):
    """
    Return non-recursive sequence of items in directory `path`

    If `path` is not a directory or getting the directory's items fails for any reason, return an
    empty sequence.
    """
    try:
        return os.listdir(path)
    except OSError:
        return ()


def file_list(path, *, extensions=(), min_age=None, max_age=None, follow_dirlinks=False):
    """
    List naturally sorted files in `path` and any subdirectories

    If `path` is not a directory, it is returned as a single item in a list
    unless `extensions` are given and they don't match.

    If `path` is any falsy value (e.g. ``""`` or `None`), return an empty
    sequence.

    Unreadable directories are excluded.

    :param str path: Path to a directory
    :param str extensions: Only include files with one of these extensions
        (matched case-insensitively) or include all files if this is falsy
    :param min_age: Exclude files that are younger than this
    :type min_age: int or float
    :param max_age: Exclude files that are older than this
    :type max_age: int or float
    :param follow_dirlinks: Whether to include the contents of symbolic links to
        directories or the links themselves

    :return: Tuple of file paths
    """
    extensions = tuple(str(e).casefold() for e in extensions)

    def ext_ok(filename):
        return not extensions or file_extension(filename).casefold() in extensions

    def age_ok(filepath, now=time.time()):
        try:
            statinfo = os.stat(filepath, follow_symlinks=False)
        except OSError:
            return False
        else:
            age = round(now - statinfo.st_atime)
            return (min_age or age) <= age <= (max_age or age)

    if not path:
        # Handle any false value, e.g. "" or `None`
        return ()
    elif not os.path.isdir(path):
        if ext_ok(path):
            return (str(path),)
        else:
            return ()

    files = []
    for root, dirnames, filenames in os.walk(path, topdown=True, followlinks=follow_dirlinks):
        for filename in sorted(filenames):
            filepath = os.path.join(root, filename)
            if ext_ok(filename) and age_ok(filepath):
                files.append(filepath)

        # Symbolic links to directories are in `dirnames`
        if not follow_dirlinks:
            # For symbolic links that point to a directory, os.walk() can only
            # include the contents (followlinks=True) or completely ignore the
            # link. Here we look for links to directories and a) include them in
            # the returned list and b), remove them from `dirnames` so os.walk()
            # doesn't recurse into them. It is important to pass topdown=True
            # for b) to work.
            for dirname in tuple(dirnames):
                # Contrary to what the docs say, this returns `True` for dead
                # and living symbolic links
                dirpath = os.path.join(root, dirname)
                if os.path.islink(dirpath):
                    del dirnames[dirnames.index(dirname)]
                    files.append(dirpath)

    return tuple(natsort.natsorted(files, key=str.casefold))


def find_name(name, path, validator=None):
    """
    Return first path (in human sort order) to `name` anywhere beneath
    `path` or `None`

    :param str name: Exact name of the file or directory
    :param str path: Base path to search for `name`

    :param validator: Callable that gets a path that ends with `name` and
        returns whether it is a match or not

        For example, you can use :func:`os.path.isfile` or :func:`os.path.isdir`
        to only find files or directories.
    """
    for root_, dirnames_, filenames_ in natsort.natsorted(
            os.walk(path, topdown=True, followlinks=False),
    ):
        names_ = natsort.natsorted(filenames_ + dirnames_)
        for name_ in names_:
            path_ = os.path.join(root_, name_)
            if name_ == name and (
                    not validator or validator(path_)
            ):
                return path_


def filter_files(path, exclude):
    """
    Return sequence of files beneath `path` with some files excluded

    If `path` is not a directory, an empty sequence is returned if it matches anything in `exclude`.

    :param exclude: Sequence of glob patterns (:class:`str`) and :class:`re.Pattern` (return value
        from :func:`re.compile`) or :class:`~.types.RegEx` objects

        Files beneath `path` are excluded from the torrent.

        Glob patterns are matched case-insensitively. For case-insensitive matching with regular
        expressions, use ``(?i:<pattern>)``.

    :raise ContentError: if anything goes wrong, e.g. a directory cannot be read
    """
    # Convert every `exclude` pattern to `re.Pattern`.
    regexs = []
    for pattern in exclude:
        if isinstance(pattern, re.Pattern):
            regexs.append(pattern)
        elif isinstance(pattern, types.RegEx):
            regexs.append(re.compile(pattern.pattern))
        elif isinstance(pattern, str):
            regexs.append(re.compile(fnmatch.translate(str(pattern))))
        else:
            raise TypeError(f'Unexpected exclude type: {pattern!r}')

    # Use torf to apply the exclude patterns instead of re-implementing. This is less work, and it
    # makes sure we get screenshots, mediainfo, etc from the same files that are in the torrent.
    try:
        torrent = torf.Torrent(path=path, exclude_regexs=regexs)
    except torf.TorfError as e:
        raise errors.ContentError(str(e)) from e
    else:
        return tuple(str(filepath) for filepath in torrent.filepaths)


def find_main_video(path):
    """
    Return main video file from `path`

    If `path` is a video file, simply return it. If `path` is a directory, pass it to
    :func:`~.fs.find_main_videos` and return the first item.

    If no video file is found, raise :class:`~.ContentError`.
    """
    main_videos = find_main_videos(path)
    if main_videos:
        return main_videos[0]
    elif os.path.isdir(path):
        raise errors.ContentError(f'No video file found: {path}')
    else:
        raise errors.ContentError(f'Not a video file: {path}')


def find_main_videos(path, exclude=()):
    """
    Return main video file paths beneath `path`

    This is a simple convenience wrapper around :func:`filter_files` and :func:`filter_main_videos`.
    """
    filtered_filepaths = filter_files(path, exclude=exclude)
    return filter_main_videos(filtered_filepaths)


def filter_main_videos(filepaths):
    """
    Exclude irrelevant file paths from `filepaths`

    Irrelevant file paths are non-video file paths (based on
    :const:`~.constants.VIDEO_FILE_EXTENSIONS`) and sample videos.

    For season packs, try to find the main episodes by file name first. If that fails, try to find
    the main video(s) exclude all video files that are smaller than 75 % of the average video file
    size.
    """
    video_filepaths = tuple(
        filepath
        for filepath in filepaths
        if file_extension(filepath).lower() in constants.VIDEO_FILE_EXTENSIONS
    )
    for filter_main_videos in (
            _filter_main_videos__episodes,
            _filter_main_videos__default,
    ):
        main_video_filepaths = filter_main_videos(video_filepaths)
        if main_video_filepaths:
            return tuple(main_video_filepaths)
    return ()

def _filter_main_videos__episodes(video_filepaths):
    # Find episodes in season pack by file name. This means we should get all episodes but ignore
    # samples and other crap even if they vary in size/duration dramatically.
    sep = r'[\._ ]'
    episode_regex = re.compile(
        r'(?:'
        + '|'.join((
            # S03E06. | .E06.
            rf'(?:S\d+|{sep})E\d+{sep}',
            # ^3x06. | .3x06. | .3.06.
            rf'(?:^|{sep})\d+[x\.]\d+{sep}',
            # .306.
            rf'({sep})\d{{3}}\1',
            # ^Episode.6. | .Episode6.
            rf'(?:^|{sep})Episode{sep}*\d+{sep}',
        ))
        + r')',
        flags=re.IGNORECASE,
    )
    matching_episodes = (
        fp
        for fp in video_filepaths
        if episode_regex.search(os.path.basename(fp))
    )
    return natsort.natsorted(matching_episodes)

def _filter_main_videos__default(video_filepaths):
    # Exclude samples, extras, etc based on average duration.
    # First, we get the size of each file and exclude any empty files and files with unknown size
    # (fs.file_size() may return `None`).
    filesizes = {
        fp: size
        for fp in video_filepaths
        if (size := file_size(fp))
    }
    # Get average and minimum file size.
    avg_size = (sum(filesizes.values()) / len(filesizes)) if filesizes else 0
    min_size = avg_size * 0.75
    return natsort.natsorted(
        filepath
        for filepath, filesize in filesizes.items()
        if filesize >= min_size
    )


# Stolen from torf-cli
# https://github.com/rndusr/torf-cli/blob/master/torfcli/_utils.py

_DOWN        = '\u2502'  # noqa: E221  # │
_DOWN_RIGHT  = '\u251C'  # noqa: E221  # ├
_RIGHT       = '\u2500'  # noqa: E221  # ─
_CORNER      = '\u2514'  # noqa: E221  # └

def format_file_tree(tree, _parents_is_last=()):
    """
    Format nested file tree sequence as indented multi-line string

    :param tree: Nested 2-tuples: The first item is the file or directory name,
        the second item is the file size for files or a tuple for directories
    """
    lines = []
    max_i = len(tree) - 1
    for i,(name,node) in enumerate(tree):
        is_last = i >= max_i

        # Assemble indentation string (`_parents_is_last` being empty means
        # this is the top node)
        indent = ''
        if _parents_is_last:
            # `_parents_is_last` holds the `is_last` values of our ancestors.
            # This lets us construct the correct indentation string: For
            # each parent, if it has any siblings below it in the directory,
            # print a vertical bar ('|') that leads to the siblings.
            # Otherwise the indentation string for that parent is empty.
            # We ignore the first/top/root node because it isn't indented.
            for parent_is_last in _parents_is_last[1:]:
                if parent_is_last:
                    indent += '  '
                else:
                    indent += f'{_DOWN} '

            # If this is the last node, use '└' to stop the line, otherwise
            # branch off with '├'.
            if is_last:
                indent += f'{_CORNER}{_RIGHT}'
            else:
                indent += f'{_DOWN_RIGHT}{_RIGHT}'

        if isinstance(node, int):
            lines.append(f'{indent}{name} ({types.Bytes(node)})')
        else:
            def sum_node_size(nodelist):
                size = 0
                for node in nodelist:
                    if isinstance(node, tuple):
                        if len(node) >= 2 and isinstance(node[1], int):
                            size += node[1]
                        else:
                            size += sum_node_size(node)
                return size

            lines.append(f'{indent}{name} ({types.Bytes(sum_node_size(node))})')
            # Descend into child node
            sub_parents_is_last = (*_parents_is_last, is_last)
            lines.append(format_file_tree(node, _parents_is_last=sub_parents_is_last))

    return '\n'.join(lines)



import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit

from .... import __project_name__, errors
from .. import widgets
from . import JobWidgetBase


class SubmitJobWidget(JobWidgetBase):

    is_interactive = False

    def setup(self):
        # Extend some error messages with additional information specific to this UI,
        # e.g. "use --option to fix this".
        self.job.signal.register('error', self._handle_error)

        # Indicate uploading acitivity.
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            format='{indicator} Uploading',
            extend_width=False,

        )
        self._activity_indicator.disable()
        self.job.signal.register('submitting', lambda: self._activity_indicator.enable())
        self.job.signal.register('submitted', lambda _: self._activity_indicator.disable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

    def _handle_error(self, error):
        if isinstance(error, errors.AnnounceUrlNotSetError):
            cmd = f'{__project_name__} set trackers.{error.tracker.name}.announce_url <URL>'
            self.job.error(f'Set it with this command: {cmd}')

        elif isinstance(error, errors.FoundDupeError):
            self.job.error('You can override the dupe check with --ignore-dupes.')

    @functools.cached_property
    def runtime_widget(self):
        return HSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._activity_indicator.active),
                    content=self._activity_indicator,
                ),
            ],
        )



"""
Entry point
"""

import asyncio
import contextlib
import sys
import threading

from ... import __homepage__, __project_name__, application_setup, application_shutdown, errors
from ...utils import update
from . import commands, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def main(args=None):
    sys.exit(_main(args))


def _main(args=None):
    cmd = None
    update_check_thread = None

    try:
        # Get CommandBase subclass. We need that to access user-provided configuration from files
        # and CLI arguments.
        cmd = commands.run(args)

        # Ugly hack.
        application_setup(cmd.config)

        if utils.is_tty():
            from .tui import TUI
            ui = TUI()
            # Find newer version in a background thread.
            update_check_thread = _UpgradeCheck(
                check_for_prerelease=(
                    cmd.config['config']['main']['check_for_prerelease']
                    or
                    update.current_is_prerelease()
                ),
            )
        else:
            # Headless mode is used when neither stdout nor stderr are TTYs, so the TUI cannot
            # interact with the user.
            # TODO: This should be in a separate package (e.g. upsies.uis.cli).
            from .headless import Headless
            ui = Headless()

        exit_code = ui.run(cmd.jobs_active)
        _log.debug('Main application exit code: %r', exit_code)

    # UI was terminated by user prematurely
    except KeyboardInterrupt as e:
        utils.print_stderr(e)
        return 1

    except (errors.UiError, errors.DependencyError, errors.ContentError) as e:
        utils.print_stderr(e)
        return 1

    except Exception as e:
        # Unexpected exception
        import traceback
        for line in traceback.format_exception(type(e), e, e.__traceback__):
            utils.print_stderr(line.strip())
        utils.print_stderr()

        # Exceptions from subprocesses should save their traceback.
        # See errors.DaemonProcessError.
        if hasattr(e, 'original_traceback'):
            utils.print_stderr(e.original_traceback)
            utils.print_stderr()

        utils.print_stderr(f'Please report the traceback above as a bug: {__homepage__}')
        return 1

    else:
        # Print last job's output to stdout for use in output redirection.
        # Ignore disabled jobs.
        if exit_code == 0 and cmd.main_job.output:
            utils.print_stdout('\n'.join(cmd.main_job.output))

        if update_check_thread and update_check_thread.message:
            utils.print_stderr(update_check_thread.message)

        return exit_code

    finally:
        if cmd is not None:
            # Cleanup cache, close HTTP session, etc.
            application_shutdown(cmd.config)


class _UpgradeCheck(threading.Thread):
    def __init__(self, *, check_for_prerelease=False):
        self._newer_version = None
        self._changelog = None
        self._check_for_prerelease = check_for_prerelease

        # daemon=True allows the application to exit if this thread is still
        # running. We don't want to wait for slow/unresponsive web servers.
        super().__init__(daemon=True)
        self.start()

    def run(self):
        with contextlib.suppress(errors.RequestError):
            self._newer_version = asyncio.run(self._get_newer_version())
            if self._newer_version:
                self._changelog = asyncio.run(self._get_changelog())

    async def _get_newer_version(self):
        if self._check_for_prerelease:
            return await update.get_newer_prerelease()
        else:
            return await update.get_newer_release()

    async def _get_changelog(self):
        if self._check_for_prerelease:
            return await update.get_latest_changelog()
        else:
            return await update.get_latest_release_changelog()

    @property
    def message(self):
        # If we found a newer version, assemble a notification message.
        if self._newer_version:
            msg = (
                '\n' + ('━' * 78) + '\n'
                + '\n'
                + '  \\O/\n'
                + f'   |     {__project_name__} {self._newer_version} has been released.\n'
                + '  / \\\n'
            )

            if self._changelog:
                msg += (
                    '\n'
                    + self._changelog
                    + '\n'
                )

            return msg



import re

sources = {
    'Blu-ray': re.compile(r'(?i:blu-?ray|bd(?:25|50|66|100))'),  # (UHD) BluRay|BD(25|50|66|100) (Remux)
    'HD-DVD': re.compile(r'(?i:hd-?dvd)'),  # HD(-)DVD
    'WEB': re.compile(r'^(?i:web)'),  # WEB(-DL|Rip)
    'HDTV': re.compile(r'(?:hd-?|)(?i:tv)'),  # HD(-)TV
    'DVD': re.compile(r'^(?i:dvd)'),  # DVD(5|9|...)
}



"""
Standardized configuration and setup howto
"""

from ... import __project_name__, constants, utils


class Howto:
    def __init__(self, tracker_cls):
        self._tracker_cls = tracker_cls
        self._section = -1

    def join(self, *sections):
        return '\n'.join(sections).strip()

    @property
    def current_section(self):
        return self._section

    @property
    def next_section(self):
        self._section += 1
        return self._section

    @property
    def introduction(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. How To Read This Howto\n'
                    '\n'
                    '   {howto.current_section}.1 Words in ALL_CAPS_AND_WITH_UNDERSCORES are placeholders.\n'
                    '   {howto.current_section}.2 Everything after "$" is a terminal command.\n'
                ),
                (
                    '{howto.next_section}. Configuration Defaults (Optional)\n'
                    '\n'
                    '    If you prefer, you can write all default values at once and then edit\n'
                    '    them in your favorite $EDITOR.\n'
                    '\n'
                    '    $ {executable} set --dump\n'
                    '    $ $EDITOR {tildify(constants.TRACKERS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.IMGHOSTS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.CLIENTS_FILEPATH)}\n'
                    '    $ $EDITOR {tildify(constants.CONFIG_FILEPATH)}\n'
                ),
            ),
            howto=self,
            executable=__project_name__,
            constants=constants,
            tildify=utils.fs.tildify_path,
        )

    @property
    def screenshots(self):
        def imghost_names(tracker):
            return ', '.join(
                tracker.TrackerConfig.defaults["image_host"].item_type.options
            )

        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Screenshots (Optional)\n'
                    '\n'
                    '   {howto.current_section}.1 Specify how many screenshots to make.\n'
                    '       $ {executable} set trackers.{tracker.name}.screenshots NUMBER_OF_SCREENSHOTS\n'
                ),
                (
                    '   {howto.current_section}.2 Specify where to host images.\n'
                    '       $ {executable} set trackers.{tracker.name}.image_host IMAGE_HOST,IMAGE_HOST,...\n'
                    '       If IMAGE_HOST is down, try the next one.\n'
                    '       Supported services: {imghost_names(tracker)}\n'
                    '\n'
                    '   {howto.current_section}.3 Configure image hosting service.\n'
                    '       $ {executable} upload-images IMAGE_HOST --help\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
            imghost_names=imghost_names,
        )

    @property
    def autoseed(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Add Uploaded Torrents To Client (Optional)\n'
                    '\n'
                    '   {howto.current_section}.1 Specify which client to add uploaded torrents to.\n'
                    '       $ {executable} set trackers.{tracker.name}.add_to CLIENT_NAME\n'
                    '       Supported clients: {client_names}\n'
                ),
                (
                    '   {howto.current_section}.2 Specify your client connection.\n'
                    '       $ {executable} set clients.CLIENT_NAME.url URL\n'
                    '       $ {executable} set clients.CLIENT_NAME.username USERNAME\n'
                    '       $ {executable} set clients.CLIENT_NAME.password PASSWORD\n'
                    '\n'
                    '{howto.next_section}. Copy Uploaded Torrents To Directory (Optional)\n'
                    '\n'
                    '   $ {executable} set trackers.{tracker.name}.copy_to /path/to/directory\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
            client_names=', '.join(utils.btclient.client_names()),
        )

    @property
    def reuse_torrents(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Reuse Existing Torrents (Optional)\n'
                    '\n'
                    '    You can skip the hashing when creating a torrent by specifying\n'
                    '    a directory path that contains the torrents you are seeding.\n'
                    '    A matching torrent is found by searching the directory recursively\n'
                    '    for a torrent with the same size and file names. If such a torrent is\n'
                    '    found, a few pieces of each file are hashed to verify the match.\n'
                    '\n'
                    '    $ {executable} set config.torrent-create.reuse_torrent_paths TORRENT_DIRECTORY\n'
                ),
            ),
            howto=self,
            executable=__project_name__,
        )

    @property
    def upload(self):
        return utils.string.evaluate_fstring(
            self.join(
                (
                    '{howto.next_section}. Upload\n'
                    '\n'
                    '   $ {executable} submit {tracker.name} --help\n'
                    '   $ {executable} submit {tracker.name} /path/to/content\n'
                ),
            ),
            howto=self,
            tracker=self._tracker_cls,
            executable=__project_name__,
        )



"""
CLI argument types

All types return normalized values and raise ValueError for invalid values.

A custom error message can be provided by raising
:class:`argparse.ArgumentTypeError`.
"""

import argparse
import functools
import os

from .. import errors, utils
from . import fs, types

natsort = utils.LazyModule(module='natsort', namespace=globals())


ArgumentTypeError = argparse.ArgumentTypeError
"""
Exception that should be raised by any callable that is passed to
:func:`argparse.ArgumentParser.add_argument` as `type` if it gets an invalid
value
"""


def comma_separated(argtype):
    """
    Multiple comma-separated values

    :param argtype: Any callable that returns a validated object for one of the
        comma-separated values or raises :class:`ValueError`, :class:`TypeError`
        or :class:`argparse.ArgumentTypeError`

    :return: Sequence of `argtype` return values
    """

    def comma_separated(value):
        values = []
        for string in str(value).split(','):
            string = string.strip()
            if string:
                try:
                    values.append(argtype(string))
                except (ValueError, TypeError) as e:
                    raise argparse.ArgumentTypeError(f'Invalid value: {string}') from e
        return values

    return comma_separated


def client(value):
    """Name of a supported BitTorrent client"""
    from . import btclient
    name = value.lower()
    if name in btclient.client_names():
        return name
    else:
        raise argparse.ArgumentTypeError(f'Unsupported client: {value}')


def content(value):
    """Existing path to release file(s)"""
    path = release(value)
    return existing_path(path)


def existing_path(value):
    """Path to existing path"""
    path = str(value)
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f'No such file or directory: {value}')
    else:
        return path


def imghost(value):
    """Name of a image hosting service from :mod:`~.utils.imghosts`"""
    from . import imghosts
    if value in imghosts.imghost_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported image hosting service: {value}')


def imghosts(value):
    """Comma-separated list of names of image hosting services from :mod:`~.utils.imghosts`"""
    names = []
    for name in value.split(','):
        name = name.strip()
        if name:
            names.append(imghost(name))
    return names


def integer(value):
    """Natural number (:class:`float` is rounded)"""
    try:
        return int(float(value))
    except (ValueError, TypeError) as e:
        raise argparse.ArgumentTypeError(f'Not an integer: {value!r}') from e


def bool_or_none(value):
    """Convert `value` to :class:`~.types.Bool` or `None` if `value` is `None`"""
    if value is None:
        return None
    else:
        try:
            return types.Bool(value)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e) from e


@functools.cache
def number_of_screenshots(*, min, max):
    """
    Return function that returns how many screenshots to make or raises
    :class:`argparse.ArgumentTypeError`

    :param int min: Minimum number of screenshots
    :param int max: Maximum number of screenshots
    """

    def number_of_screenshots(value):
        """How many screenshots to make within allowed range"""
        try:
            return types.Integer(value, min=min, max=max)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e) from e

    return number_of_screenshots


@functools.cache
def files_with_extension(extension, *, allow_no_hits=True):
    """
    Return function that recursively searches a directory for files with
    `extension`

    If the returned function gets a file path with the wanted extension, it is
    simply returned.

    :param str extension: Wanted file name extension (e.g. "png")
    :param bool allow_no_hits: Whether :exc:`argparse.ArgumentTypeError` is
        raised if no matching files are found
    """

    def is_match(filepath):
        if fs.file_extension(filepath).casefold() == extension.casefold():
            try:
                fs.assert_file_readable(filepath)
            except errors.ContentError as e:
                raise argparse.ArgumentTypeError(e) from e
            else:
                return True
        return False

    def files_with_extension(value):
        matching_files = []

        if os.path.isdir(value):
            for dirpath, _dirnames, filenames in os.walk(value):
                for filename in filenames:
                    filepath = os.path.join(dirpath, filename)
                    if is_match(filepath):
                        matching_files.append(filepath)

            if not matching_files and not allow_no_hits:
                raise argparse.ArgumentTypeError(f'{value}: No {extension} files found')

        else:
            if os.path.exists(value):
                if is_match(value):
                    matching_files = (value,)
                else:
                    msg = f'Expected file extension {extension}'
                    if ext := fs.file_extension(value):
                        msg += f', not {ext}'
                    msg += f': {value}'
                    raise argparse.ArgumentTypeError(msg)

            if not matching_files and not allow_no_hits:
                raise argparse.ArgumentTypeError(f'{value}: Not a {extension} file')

        return tuple(natsort.natsorted(
            matching_files,
            key=lambda filepath: fs.basename(filepath).casefold(),
        ))

    return files_with_extension


def option(value):
    """Name of a configuration option"""
    from .. import defaults
    if value in defaults.option_paths():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unknown option: {value}')


@functools.cache
def one_of(values):
    """
    Return function that returns an item of `values` or raises
    :class:`argparse.ArgumentTypeError`

    :param values: Allowed values
    """
    values = tuple(values)

    def one_of_values(value):
        if value in values:
            return value
        else:
            raise argparse.ArgumentTypeError(f'Invalid value: {value}')

    return one_of_values


def regex(value):
    """:class:`re.Pattern` object"""
    try:
        return types.RegEx(value)
    except ValueError as e:
        raise argparse.ArgumentTypeError(e) from e


def release(value):
    """Same as :func:`content`, but doesn't have to exist"""
    from .. import errors
    from . import predbs
    path = str(value)
    try:
        predbs.assert_not_abbreviated_filename(path)
    except errors.SceneAbbreviatedFilenameError as e:
        raise argparse.ArgumentTypeError(e) from e
    else:
        return path


def predb_name(value):
    """Name of a scene release database from :mod:`~.utils.predbs`"""
    from . import predbs
    if value in predbs.predb_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported scene release database: {value}')


def predb(value):
    """
    :class:`~.PredbApiBase` instance from a corresponding
    :attr:`~.PredbApiBase.name`
    """
    from . import predbs
    try:
        return predbs.predb(value.lower())
    except ValueError as e:
        raise argparse.ArgumentTypeError(e) from e


def timestamp(value):
    """Turn `value` into :class:`types.Timestamp`"""
    try:
        return types.Timestamp.from_string(value)
    except (ValueError, TypeError) as e:
        raise argparse.ArgumentTypeError(e) from e


def tracker(value):
    """Name of a tracker from :mod:`~.trackers`"""
    from .. import trackers
    if value in trackers.tracker_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported tracker: {value}')


def webdb(value):
    """Name of a movie/series database from :mod:`~.webdbs`"""
    from . import webdbs
    if value in webdbs.webdb_names():
        return value.lower()
    else:
        raise argparse.ArgumentTypeError(f'Unsupported database: {value}')


@functools.cache
def webdb_id(webdb_name):
    """
    Return function that finds a web DB ID in a string, e.g. an URL

    :param str webdb_name: Name of a web DB, e.g. "imdb"

    The returned function takes any object and passes it to
    :meth:`~.WebDbApiBase.get_id_from_text`.
    """

    from . import webdbs
    db = webdbs.webdb(webdb_name)

    def webdb_id(value):
        id = db.get_id_from_text(str(value))
        if id:
            return id
        else:
            raise argparse.ArgumentTypeError(f'Invalid {db.name} ID: {value}')

    return webdb_id



import functools

from prompt_toolkit.layout.containers import DynamicContainer, HSplit, VSplit

from .... import utils
from .. import widgets
from . import JobWidgetBase


class BdinfoJobWidget(JobWidgetBase):

    is_interactive = False

    def setup(self):
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            extend_width=False,
        )
        self._bdinfo_progress_widgets = {}

        self.job.siblings['playlists'].signal.register('playlists_selected', self.handle_playlists_selected)
        self.job.signal.register('bdinfo_progress', self.handle_bdinfo_progress)
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

    def handle_playlists_selected(self, discpath, playlists):
        # Add progress info widgets for each playlist.
        for playlist in playlists:
            label = playlist.label
            if label not in self._bdinfo_progress_widgets:
                self._bdinfo_progress_widgets[label] = BdinfoProgressWidget(label)

    def handle_bdinfo_progress(self, progress):
        self._bdinfo_progress_widgets[progress.playlist.label].update(progress)
        self.invalidate()

    @functools.cached_property
    def runtime_widget(self):
        return DynamicContainer(self.get_runtime_widgets)

    def get_runtime_widgets(self):
        if not self._bdinfo_progress_widgets:
            # Show activity indicator if no BDInfo report generation has started yet.
            self._activity_indicator.enable()
            return self._activity_indicator
        else:
            self._activity_indicator.disable()
            return HSplit(self._bdinfo_progress_widgets.values())


class BdinfoProgressWidget:

    _instances = []

    def __new__(cls, *_args, **_kwargs):
        self = super().__new__(cls)
        cls._instances.append(self)
        return self

    def __init__(self, label):
        self._label = str(label)
        self._activity_indicator = widgets.ActivityIndicator()
        self._bar = widgets.ProgressBar(width=30)
        self._label_widget = widgets.TextField(text=lambda: self.label_padded, extend_width=False)
        self._time_elapsed = utils.types.Timestamp(0)
        self._time_elapsed_widget = widgets.TextField(lambda: str(self._time_elapsed), extend_width=False)
        self._time_remaining = utils.types.Timestamp(0)
        self._time_remaining_widget = widgets.TextField(lambda: str(self._time_remaining), extend_width=False)
        self._container = VSplit(
            children=[
                self._label_widget,
                DynamicContainer(self._get_progress_widgets),
            ],
            style='class:info',
            width=lambda: self.label_width + len(' 0:00:10                                0:01:30'),
        )

    def _get_progress_widgets(self):
        if self.is_waiting_for_first_progress:
            # Show activity indicator if no bdinfo is currently generated and this instance is
            # next in line.
            return VSplit([
                widgets.hspacer,
                self._activity_indicator,
            ])

        elif 0 < self._bar.percent < 100:
            # Show progress if report is being generated.
            return VSplit([
                widgets.hspacer,
                self._time_elapsed_widget,
                widgets.hspacer,
                self._bar,
                widgets.hspacer,
                self._time_remaining_widget,
            ])
        elif self._bar.percent >= 100:
            # Report generation is done.
            return VSplit([
                widgets.hspacer,
                widgets.TextField('✔', extend_width=False),
            ])
        else:
            return widgets.hspacer

    @property
    def is_waiting_for_first_progress(self):
        prev_instance = None
        for instance in type(self)._instances:
            if (
                    instance is self
                    # Report generation hasn't started/finished yet.
                    and instance.percent <= 0
                    # Previous instance finished or this is the first instance.
                    and (
                        prev_instance is None
                        or prev_instance.percent >= 100
                    )
            ):
                self._activity_indicator.enable()
                return True

            prev_instance = instance

        return False

    def update(self, progress):
        self.percent = progress.percent
        self.time_elapsed = progress.time_elapsed
        self.time_remaining = progress.time_remaining

    @property
    def label(self):
        return self._label

    @property
    def label_padded(self):
        return self.label.rjust(self.label_width)

    @property
    def label_width(self):
        return max(
            len(instance.label)
            for instance in type(self)._instances
        )

    @property
    def percent(self):
        return self._bar.percent

    @percent.setter
    def percent(self, percent):
        self._bar.percent = percent
        self._activity_indicator.disable()

    @property
    def time_elapsed(self):
        return self._time_elapsed

    @time_elapsed.setter
    def time_elapsed(self, time_elapsed):
        self._time_elapsed = utils.types.Timestamp(time_elapsed)

    @property
    def time_remaining(self):
        return self._time_remaining

    @time_remaining.setter
    def time_remaining(self, time_remaining):
        self._time_remaining = utils.types.Timestamp(time_remaining)

    def __pt_container__(self):
        return self._container
